#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

//#include "header\font\BGFont.h"
//#include "header\font\BitMapFont.h"

#include "lib.h"
#include "syscall.h"
#include "resource.h"

#include <stdlib.h>

//Mode0

//^CgAGfBO
//BG2 
//BG3 

//Q[
//BG2 Xe[W
//BG3 
//SPR LN^

//---------------------------------------------------------------------------
#define STATE_INIT 						1		//is
#define STATE_TITLE						2
#define STATE_OPENING					3
#define STATE_GAME						4
#define STATE_ENDING					5

#define BG_SIZE							16
#define BG_NO_WHITE						20
#define BG_NO_INVISIBLE					0

#define FIELD_CX						120
#define FIELD_CY						60
#define FIELD_SCROLLPANEL_SPEED			2.0

#define FIELD_BG_NONE					0
#define FIELD_BG_HIDEPATH				2
#define FIELD_BG_UNVISIBLE				4
#define FIELD_BG_BLOCK					6
#define FIELD_BG_BAR					8
#define FIELD_BG_SCROLL_L				10
#define FIELD_BG_SCROLL_R				12
#define FIELD_BG_SPIKE					14
#define FIELD_BG_SLIP					16
#define FIELD_ITEM_BORDER               127 //ACe`FbNp

#define FIELD_ITEM_MAX					19  //ACe̍viӂn`[nj
#define FIELD_BG_ITEM_POWERUP			256 //p[Abv
#define FIELD_BG_ITEM_FUJI				258 //ӂn
#define FIELD_BG_ITEM_BUSHI				260
#define FIELD_BG_ITEM_APPLE				262
#define FIELD_BG_ITEM_V					264
#define FIELD_BG_ITEM_TAKA				266 //n
#define FIELD_BG_ITEM_SHUOLDER			268
#define FIELD_BG_ITEM_DAGGER			270
#define FIELD_BG_ITEM_KATAKATA			272
#define FIELD_BG_ITEM_NASU				274 //Ȃn
#define FIELD_BG_ITEM_BONUS				276
#define FIELD_BG_ITEM_NURSE				278
#define FIELD_BG_ITEM_NAZUNA			280
#define FIELD_BG_ITEM_GAMEHELL			282 //[n
#define FIELD_BG_ITEM_GUNDAM			284
#define FIELD_BG_ITEM_POED				286
#define FIELD_BG_ITEM_MILESTONE			320
#define FIELD_BG_ITEM_1YEN				322
#define FIELD_BG_ITEM_TRIANGLE			324
#define FIELD_BG_ITEM_OMEGA				326 //B
#define FIELD_BG_STARTPOINT				999 //Jnn_

#define PLAYER_SPEED					2.0
#define PLAYER_GRD_ACCRATIO				0.04
#define PLAYER_AIR_ACCRATIO				0.01
#define PLAYER_JUMP						-4.0
#define PLAYER_GRAVITY					0.2
#define PLAYER_FALL_SPEEDMAX			4.0
#define PLAYER_VIEW_DIRECTION_OFFSET	30.0
#define PLAYER_WAIT_TIMER_INTERVAL		10
#define PLAYER_LIFE_RATIO				400
#define PLAYER_LIFE_MAX					(PLAYER_LIFE_RATIO * 3)
#define PLAYER_MUTEKI_INTERVAL			50
#define PLAYER_START_WAIT_INTERVAL		50
#define PLAYER_STATE_START				0
#define PLAYER_STATE_NORMAL				1
#define PLAYER_STATE_ITEMGET			2
#define PLAYER_STATE_MUTEKI				3
#define PLAYER_STATE_DEAD				4

#define CR								0x0d
#define LF								0x0a
//---------------------------------------------------------------------------
typedef struct {
	f32      x;
	f32      y;
} ST_VEC2D;

typedef struct {
	s16      LineCnt;
	u16*     pChrSrc;
	u16*     pChrDst;
	u16*     pChrDstStart;
	u16*     pChrDstEnd;
	u16*     pMapDst;
	u16*     pMapDstStart;
	u16*     pMapDstEnd;
	u16      MapSrcCnt;
	u16*     pVfos;
	u8       VfosCnt;
	u8       Wait;
	u8       WaitCnt;
	u8       DelChr;
} ST_SCROLL;

typedef struct {
	u16      Field[FIELD_CX][FIELD_CY];
	u16      Timer;
} ST_STAGE;

typedef struct {
	u8       State;							//
	ST_VEC2D Position;						//W
	ST_VEC2D Speed;							//x
	int      Direction;						//
	u16      Life;							//Ct|Cg
	s8       JumpCnt;						//Wv
	u16      ItemGet;						//擾ACe
	u16      Timer;							//^C}
	u16      WaitTimer;
	u16      Key;

	bool     ItemGetFlag[FIELD_ITEM_MAX];
	u16      PlayTime;
	u8       JumpMax;						//Wv
	ST_VEC2D View;
} ST_PLAYER;
//---------------------------------------------------------------------------
//lib.c
extern ST_TIMER Timer;

//---------------------------------------------------------------------------
ST_BG     BG1;
ST_BG     BG2;
ST_BG     BG3;
ST_STAGE  Stage;
ST_PLAYER Player;

u8 State;									//is󋵃tO
//---------------------------------------------------------------------------
//Q[
void StateInit();
void StateTitle();
void StateOpening();
void StateGame();
void StateEnding();

//
void InitWS();
void InitBG();
void InitKey();
void InitSprite();
void IRQUserHandler();
void InitGame();
void InitField();
void InitPlayer();

//tB[h
void FieldLoadData(u8* fp, u16 size);
bool FieldGetStartPoint();
u8   FieldGetArea(u16 x, u16 y);
void FieldEraseArea(u16 x, u16 y);
bool FieldIsWall(u16 x, u16 y);
bool FieldIsItem(u16 x, u16 y);
bool FieldIsRidable(u16 x, u16 y);
bool FieldIsSpike(u16 x, u16 y);
bool FieldIsItemGetable(u16 x, u16 y);
bool FieldIsOpenSecret();
void FieldMove();
void FieldDraw();

//vC[
void PlayerMove();
void PlayerMoveNormal();
void PlayerMoveItemGet();
void PlayerMoveByInput();
void PlayerCheckCollision();
u8   PlayerGetOnField();
bool PlayerOnWall();
bool PlayerIsFallable();
bool PlayerIsUpperWallBoth();
bool PlayerIsUpperWall();
bool PlayerIsLeftWall();
bool PlayerIsRightWall();
void PlayerNormalizeToRight();
void PlayerNormalizeToLeft();
void PlayerNormalizeToUpper();
u16  PlayerToFieldX();
u16  PlayerToFieldY();
u16  PlayerToFieldOfsX();
u16  PlayerToFieldOfsY();
void PlayerDraw();
bool PlayerIsPressEnter();
bool PlayerIsPushEnter();
bool PlayerIsPressDown();
bool PlayerIsPressLeft();
bool PlayerIsPressRight();
void PlayerViewSetPosition(ST_VEC2D* p);

//XN[
void ScrollBG(ST_SCROLL* p);
bool ScrollBGSetNextLine(ST_SCROLL* p);
void ScrollBGSetNextChr (ST_SCROLL* p);
void ScrollBGSetNextMap (ST_SCROLL* p);

//̑
bool KeyWait(u16 key);
void BGDraw16(u8 x, u8 y, u16 tile);

//---------------------------------------------------------------------------
int main()
{
	State = STATE_INIT;

	for(;;)
	{
		switch(State)
		{
			case STATE_INIT:    StateInit();    break;
			case STATE_TITLE:   StateTitle();   break;
			case STATE_OPENING: StateOpening(); break;
			case STATE_GAME:    StateGame();    break;
			case STATE_ENDING:  StateEnding();  break;

			default:
				DebugMessage((u8*)"Err: Main State");
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void StateInit()
{
	InitWS();
	InitBG();
	InitKey();
	InitSprite();

	InitGame();

	State = STATE_TITLE;
}
//---------------------------------------------------------------------------
void InitWS()
{
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;
}
//---------------------------------------------------------------------------
void InitBG()
{

	BG1.Chr = 1;
	BG1.Map = 25;
	BG1.Col = 0;

	BG2.Chr = 0;
	BG2.Map = 26;
	BG2.Col = 0;

	BG3.Chr = 0;
	BG3.Map = 27;
	BG3.Col = 0;

	REG_BG1CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG1.Chr) | BG_MAPBASE(BG1.Map) | 1);

	REG_BG2CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG2.Chr) | BG_MAPBASE(BG2.Map) | 2);

	REG_BG3CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG3.Chr) | BG_MAPBASE(BG3.Map) | 3);

	BGSetPal16(0, (u16*)&pal_img_bin);
	BGSetChrData(BG1.Chr, (u16*)&bg_title_img_bin, BG_TITLE_IMG_SIZE);
	BGSetChrData(BG2.Chr, (u16*)&bg_stage_img_bin, BG_STAGE_IMG_SIZE);
}
//---------------------------------------------------------------------------
void InitSprite()
{
	SpriteInit();

	SpriteSetPal16((u16*)&pal_img_bin);
	SpriteSetData ((u16*)&spr_chr_img_bin, SPR_CHR_IMG_SIZE);
}
//---------------------------------------------------------------------------
void InitKey()
{
	KeyInit();

	REG_INTERUPT = (u32)IRQUserHandler;
	REG_IE       = IRQ_BIT_TIMER0;
	REG_IME      = IRQ_MASTER_ON;
	REG_TM0D     = 65536 - 2621;			// (((16*1024*1024) / 64) * 10) / 1000
	REG_TM0CNT   = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;
}
//---------------------------------------------------------------------------
void IRQUserHandler()
{
	Timer.clockTick += TIMER_INTMS;
	REG_IF = IRQ_BIT_TIMER0;
}
//---------------------------------------------------------------------------
void StateTitle()
{
	BGDrawOrder(BG1.Map);
	BGDrawCls  (BG2.Map, BG_NO_INVISIBLE);
	BGDrawCls  (BG3.Map, BG_NO_WHITE);

	SetMode(MODE_0 | BG1_ENABLE | BG2_ENABLE | BG3_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);
	WaitKey(KEY_A);

	State = STATE_OPENING;
}
//---------------------------------------------------------------------------
void StateOpening()
{
	ST_SCROLL opScroll;

	opScroll.LineCnt      = OPENING_IMG_CY + SCREEN_CY;
	opScroll.pChrSrc      = (u16*)&bg_opening_img_bin;
	opScroll.pChrDst      = MEM_BG_CHR(BG1.Chr + 0);
	opScroll.pChrDstStart = MEM_BG_CHR(BG1.Chr + 0);
	opScroll.pChrDstEnd   = MEM_BG_CHR(BG1.Chr + 2);
	opScroll.pMapDst      = MEM_BG_MAP(BG1.Map + 0) + BG_WORLD_SCREEN_CX * (SCREEN_CY / BG_CY);
	opScroll.pMapDstStart = MEM_BG_MAP(BG1.Map + 0);
	opScroll.pMapDstEnd   = MEM_BG_MAP(BG1.Map + 1);
	opScroll.MapSrcCnt    = 0;
	opScroll.pVfos        = (u16*)&REG_BG1VOFS;
	opScroll.VfosCnt      = 256 - 8;
	opScroll.Wait         = 3;
	opScroll.WaitCnt      = 0;
	opScroll.DelChr       = 0;

	BGDrawCls(BG1.Map, BG_NO_INVISIBLE);
	ScrollBG(&opScroll);
	BGDrawCls(BG1.Map, BG_NO_INVISIBLE);

	State = STATE_GAME;
}
//---------------------------------------------------------------------------
void StateEnding()
{
	ST_SCROLL edScroll;

	edScroll.LineCnt      = ENDING_IMG_CY + SCREEN_CY;
	edScroll.pChrSrc      = (u16*)&bg_ending_img_bin;
	edScroll.pChrDst      = MEM_BG_CHR(BG1.Chr + 0);
	edScroll.pChrDstStart = MEM_BG_CHR(BG1.Chr + 0);
	edScroll.pChrDstEnd   = MEM_BG_CHR(BG1.Chr + 2);
	edScroll.pMapDst      = MEM_BG_MAP(BG1.Map + 0) + BG_WORLD_SCREEN_CX * (SCREEN_CY / BG_CY);
	edScroll.pMapDstStart = MEM_BG_MAP(BG1.Map + 0);
	edScroll.pMapDstEnd   = MEM_BG_MAP(BG1.Map + 1);
	edScroll.MapSrcCnt    = 0;
	edScroll.pVfos        = (u16*)&REG_BG1VOFS;
	edScroll.VfosCnt      = 256 - 8;
	edScroll.Wait         = 3;
	edScroll.WaitCnt      = 0;
	edScroll.DelChr       = 0;

	BGDrawCls(BG1.Map, BG_NO_INVISIBLE);
	ScrollBG(&edScroll);
	BGDrawCls(BG1.Map, BG_NO_INVISIBLE);

	State = STATE_INIT;
}
//---------------------------------------------------------------------------
void ScrollBG(ST_SCROLL* p)
{
	while(p->LineCnt > 0)
	{
		WaitForVsync();

		if( KeyWait(KEY_A) == TRUE )
		{
			goto EndFunc;
		}

		if(p->WaitCnt++ < p->Wait)
		{
			continue;
		}
		p->WaitCnt = 0;

		if( ScrollBGSetNextLine(p) == TRUE )
		{
			ScrollBGSetNextChr(p);
			ScrollBGSetNextMap(p);
		}
		p->LineCnt--;
	}

EndFunc:
	p->pVfos = 0;
}
//---------------------------------------------------------------------------
bool ScrollBGSetNextLine(ST_SCROLL* p)
{
	p->VfosCnt = (p->VfosCnt + 1) % BG_WORLD_SCREEN_WCY;
	*p->pVfos  = (p->VfosCnt);

	return p->VfosCnt % BG_CY == 0 ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
void ScrollBGSetNextChr(ST_SCROLL* p)
{
	s16 i;
	u16 loopEnd = BG_SCREEN_CX * BG_CHR_BYTE_SIZE / 2;

	//Ō̕̕\ȍ~́A󔒕\ɂ܂
	if(p->LineCnt >= SCREEN_CY)
	{
		for(i=0; i<loopEnd; i++)
		{
			*p->pChrDst++ = *p->pChrSrc++;
		}
	}
	else
	{
		for(i=0; i<loopEnd; i++)
		{
			*p->pChrDst++ = 0x0000;
		}
	}

	if(p->pChrDst >= p->pChrDstEnd)
	{
		p->pChrDst = p->pChrDstStart;
	}
}
//---------------------------------------------------------------------------
void ScrollBGSetNextMap(ST_SCROLL* p)
{
	s16 i;

	for(i=0; i<BG_SCREEN_CX; i++)
	{
		*p->pMapDst++ = p->MapSrcCnt++;
	}
	p->pMapDst += 2;

	if(p->pMapDst >= p->pMapDstEnd)
	{
		p->pMapDst = p->pMapDstStart;
	}

	if(p->MapSrcCnt >= BG_MAP_MAX_CNT)
	{
		p->MapSrcCnt = 0;
	}
}
//---------------------------------------------------------------------------
//VSYNC҂Ăs邱
bool KeyWait(u16 key)
{
	return KeyInput() & key ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
void StateGame()
{
	WaitForVsync();

	FieldDraw();
	PlayerDraw();

	FieldMove();
	PlayerMove();
}
//---------------------------------------------------------------------------
void InitGame()
{
	InitField();
	FieldLoadData((u8*)&dat_field_img_bin, DAT_FIELD_IMG_SIZE);

	InitPlayer();

	FieldGetStartPoint();
	PlayerViewSetPosition(&Player.Position);
}
//---------------------------------------------------------------------------
void InitPlayer()
{
	ST_PLAYER* p = &Player;

	p->State      = PLAYER_STATE_START;
	p->ItemGet    = 0;
	p->Timer      = 0;
	p->WaitTimer  = 0;
	p->Life       = PLAYER_LIFE_MAX;
	p->JumpCnt    = 0;
	p->Speed.x    = 0.0;
	p->Speed.y    = 0.0;
	p->Position.x = 0;
	p->Position.y = 0;
	p->Direction  = 0;
	p->Key        = 0x0000;

	s16 i;
	for(i=0; i<FIELD_ITEM_MAX; i++)
	{
		p->ItemGetFlag[i] = FALSE;
	}
	p->JumpMax  = 0;
	p->PlayTime = 0;
}
//---------------------------------------------------------------------------
void InitField()
{
	ST_STAGE* p = &Stage;
	s16 x, y;

	for(y=0; y<FIELD_CY; y++)
	{
		for(x=0; x<FIELD_CX; x++)
		{
			p->Field[x][y] = FIELD_BG_NONE;
		}
	}

	p->Timer = 0;
}
//---------------------------------------------------------------------------
void FieldLoadData(u8* fp, u16 size)
{
	u8* pEndLoop = fp + size;
	u8  ch;
	u16 bg;
	s16 x, y;

	x  = 0;
	y  = 0;
	bg = 0;
	ch = 0;

	while(fp < pEndLoop)
	{
		ch = *fp++;

		if(ch == CR)
		{
			x = 0; y++;

			if(*fp == LF)
			{
				//LF΂ǂ݂܂
				fp++;
			}
			continue;
		}

		switch(ch)
		{
		case ' ': bg = FIELD_BG_NONE;           break;
		case 'H': bg = FIELD_BG_HIDEPATH;       break;
		case 'U': bg = FIELD_BG_UNVISIBLE;      break;
		case 'B': bg = FIELD_BG_BLOCK;          break;
		case '~': bg = FIELD_BG_BAR;            break;
		case '<': bg = FIELD_BG_SCROLL_L;       break;
		case '>': bg = FIELD_BG_SCROLL_R;       break;
		case '*': bg = FIELD_BG_SPIKE;          break;
		case 'I': bg = FIELD_BG_SLIP;           break;
		case 'P': bg = FIELD_BG_ITEM_POWERUP;   break;
		case 'a': bg = FIELD_BG_ITEM_FUJI;      break;
		case 'b': bg = FIELD_BG_ITEM_BUSHI;     break;
		case 'c': bg = FIELD_BG_ITEM_APPLE;     break;
		case 'd': bg = FIELD_BG_ITEM_V;         break;
		case 'e': bg = FIELD_BG_ITEM_TAKA;      break;
		case 'f': bg = FIELD_BG_ITEM_SHUOLDER;  break;
		case 'g': bg = FIELD_BG_ITEM_DAGGER;    break;
		case 'h': bg = FIELD_BG_ITEM_KATAKATA;  break;
		case 'i': bg = FIELD_BG_ITEM_NASU;      break;
		case 'j': bg = FIELD_BG_ITEM_BONUS;     break;
		case 'k': bg = FIELD_BG_ITEM_NURSE;     break;
		case 'l': bg = FIELD_BG_ITEM_NAZUNA;    break;
		case 'm': bg = FIELD_BG_ITEM_GAMEHELL;  break;
		case 'n': bg = FIELD_BG_ITEM_GUNDAM;    break;
		case 'o': bg = FIELD_BG_ITEM_POED;      break;
		case 'p': bg = FIELD_BG_ITEM_MILESTONE; break;
		case 'q': bg = FIELD_BG_ITEM_1YEN;      break;
		case 'r': bg = FIELD_BG_ITEM_TRIANGLE;  break;
		case 'z': bg = FIELD_BG_ITEM_OMEGA;     break;
		case '@': bg = FIELD_BG_STARTPOINT;     break;
		default :
			DebugMessage((u8*)"Err: FieldLoadData ch");
			break;
		}

		Stage.Field[x][y] = bg;
		x++;
	}

}
//---------------------------------------------------------------------------
bool FieldGetStartPoint()
{
	ST_STAGE*  s = &Stage;
	ST_PLAYER* p = &Player;
	s16 x, y;

	for(y=0; y<FIELD_CY; y++)
	{
		for(x=0; x<FIELD_CX; x++)
		{
			if(s->Field[x][y] == FIELD_BG_STARTPOINT)
			{
				p->Position.x = (f32)(x * BG_SIZE);
				p->Position.y = (f32)(y * BG_SIZE);
				FieldEraseArea(x, y);

				return TRUE;
			}
		}
	}

	return FALSE;
}
//---------------------------------------------------------------------------
void FieldEraseArea(u16 x, u16 y)
{
	ST_STAGE* p = &Stage;
	p->Field[x][y] = FIELD_BG_NONE;
}
//---------------------------------------------------------------------------
bool FieldIsWall(u16 x, u16 y)
{
	ST_STAGE* p = &Stage;

	if( p->Field[x][y] != FIELD_BG_NONE     && 
		p->Field[x][y] != FIELD_BG_HIDEPATH && 
		p->Field[x][y] != FIELD_BG_BAR      &&
		FieldIsItem(x,y) == FALSE )
	{
		return TRUE;
	}

	return FALSE;
}
//---------------------------------------------------------------------------
bool FieldIsItem(u16 x, u16 y)
{
	ST_STAGE* p = &Stage;

	if( p->Field[x][y] <  FIELD_ITEM_BORDER || 
		p->Field[x][y] == FIELD_BG_STARTPOINT )
	{
		return FALSE;
	}
	return TRUE;
}
//---------------------------------------------------------------------------
bool FieldIsRidable(u16 x, u16 y)
{
	ST_STAGE* p = &Stage;

	if( p->Field[x][y] != FIELD_BG_NONE     &&
		p->Field[x][y] != FIELD_BG_HIDEPATH &&
		FieldIsItem(x,y) == FALSE )
	{
		return TRUE;
	}
	return FALSE;
}
//---------------------------------------------------------------------------
bool FieldIsSpike(u16 x, u16 y)
{
	ST_STAGE* p = &Stage;
	return p->Field[x][y] == FIELD_BG_SPIKE ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
bool FieldIsItemGetable(u16 x, u16 y)
{
	ST_STAGE* p = &Stage;

	if( !FieldIsItem(x, y) == TRUE )
	{
		return FALSE;
	}

	if( p->Field[x][y] == FIELD_BG_ITEM_OMEGA && FieldIsOpenSecret() )
	{
		return FALSE;
	}

	return TRUE;
}
//---------------------------------------------------------------------------
bool FieldIsOpenSecret()
{
	ST_PLAYER* p = &Player;
	return p->ItemGet < 17 ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
u8 FieldGetArea(u16 x, u16 y)
{
	ST_STAGE* p = &Stage;
	return p->Field[x][y];
}
//---------------------------------------------------------------------------
void FieldMove()
{
	ST_STAGE* p = &Stage;
	p->Timer++;
}
//---------------------------------------------------------------------------
void FieldDraw()
{
	ST_STAGE*  s = &Stage;
	ST_PLAYER* p = &Player;
	s16 x, y;
	s16 tx, ty;

//̏ꂵ̂R[fBOiOOG
start:

//	WaitForVsync();

	tx = ((s16)p->View.x / BG_SIZE) - 8;
	ty = ((s16)p->View.y / BG_SIZE) - 6;

	for(y=0; y<BG_SCREEN_CY/2; y++)
	{
		for(x=0; x<BG_SCREEN_CX/2; x++)
		{
			BGDraw16(x, y, s->Field[x + tx][y + ty]);
		}
	}

in:
	if( KeyWaitTimer() == TRUE )
	{
		goto in;
	}
	if( KeyRepeat() == TRUE )		//keyGet֐ĂԑOɕKsȂ
	{
		goto in;
	}

	u8 key = KeyInput();
	if(key == 0x00)
	{
		goto in;
	}

	if(key & KEY_UP   ) p->View.y--;
	if(key & KEY_DOWN ) p->View.y++;
	if(key & KEY_LEFT ) p->View.x--;
	if(key & KEY_RIGHT) p->View.x++;

goto start;


/*
	Vec2D v = gamemain.view.getPosition();
	int ofs_x = BG_SIZE - (cast(int)v.x) % BG_SIZE;
	int ofs_y = BG_SIZE - (cast(int)v.y) % BG_SIZE;
	for(int xx = -12;xx < 13;xx++){
		for(int yy = -8;yy < 9;yy++){
			int fx = xx + cast(int)(v.x / BG_SIZE);
			int fy = yy + cast(int)(v.y / BG_SIZE);
			if(fx < 0 || fy < 0 || fx >= FIELD_X_MAX || fy >= FIELD_Y_MAX) continue;
			
			int anime = (timer / 10) % 4;
			if(isItem(fx,fy)) anime = 0;
			
			if(isOpenSecret() && field[fy * FIELD_X_MAX + fx] == FIELD_ITEM_OMEGA) continue;
			Hell_draw("field" ,
				(xx + 12)* BG_SIZE + ofs_x + GRAPHIC_OFFSET_X,
				(yy + 8) * BG_SIZE + ofs_y + GRAPHIC_OFFSET_Y ,
				field[fy * FIELD_X_MAX + fx] * 16 , anime * 16 , 16,16);
				// fobOp
			//Hell_drawFont(std.string.toString(field[fy * FIELD_X_MAX + fx]),
			//	(xx + 12)* BG_SIZE + ofs_x + GRAPHIC_OFFSET_X,
			//	(yy + 8) * BG_SIZE + ofs_y + GRAPHIC_OFFSET_Y);
		}
	}
*/
}
//---------------------------------------------------------------------------
void PlayerMove()
{
	ST_PLAYER* p = &Player;

	switch(p->State)
	{
	//Jn
	case PLAYER_STATE_START:

		p->WaitTimer++;
		if(p->WaitTimer > PLAYER_START_WAIT_INTERVAL)
		{
			p->State = PLAYER_STATE_NORMAL;
		}
		return;

	//ʏ
	case PLAYER_STATE_NORMAL:
		PlayerMoveByInput();
		PlayerMoveNormal();

		//Ct
		if(p->Life < PLAYER_LIFE_MAX)
		{
			p->Life++;
		}
		return;

	//ACeǁ[!
	case PLAYER_STATE_ITEMGET:
		PlayerMoveItemGet();

		//NA`FbN TODO
//		if( PlayerIsGameClear() == TRUE && 
//			p->State            != PLAYER_STATE_ITEMGET )
//		{
		  //  gamemain.setMsg(GameMain.MSG_REQ_ENDING); //TODO
//		}
		return;

	//G
	case PLAYER_STATE_MUTEKI:
		PlayerMoveByInput();
		PlayerMoveNormal();

		p->WaitTimer++;
		if(p->WaitTimer > PLAYER_MUTEKI_INTERVAL)
		{
			p->State = PLAYER_STATE_NORMAL;
		}
		return;

	//S
	case PLAYER_STATE_DEAD:
		PlayerMoveNormal();

//		if(isPushEnter()) gamemain.setMsg(GameState.MSG_REQ_TITLE);
		return;

	default:
		DebugMessage((u8*)"Err: PlayerMove State");
		return;

	}
}
//---------------------------------------------------------------------------
void PlayerMoveNormal()
{
	ST_PLAYER* p = &Player;

	p->Timer++;
	p->PlayTime = (p->Timer / 50);

	//--------------------------------------------------------------------
	//ړ
	p->Speed.y    += PLAYER_GRAVITY;
	p->Position.x += p->Speed.x;
	p->Position.y += p->Speed.y;

	if(p->Speed.y > PLAYER_FALL_SPEEDMAX)
	{
		p->Speed.y = PLAYER_FALL_SPEEDMAX;
	}

	if(p->State == PLAYER_STATE_NORMAL)
	{
		PlayerCheckCollision();
	}

	//--------------------------------------------------------------------
	//n
	if( PlayerOnWall() == TRUE && p->Speed.y >= 0 )
	{
		if( PlayerIsPressEnter() != TRUE &&
		    PlayerIsPressDown()  != TRUE &&
		    PlayerIsFallable()   != TRUE )
		{
			if(p->Speed.y > 0)
			{
				p->Speed.y = 0;
			}
			p->Position.y = BG_SIZE * PlayerToFieldY();
			p->JumpCnt    = 0;
		}
	}

	//--------------------------------------------------------------------
	//蔻
	bool hitLeft  = FALSE;
	bool hitRight = FALSE;
	bool hitUpper = FALSE;

	if(PlayerIsLeftWall()  && p->Speed.x <  0) hitLeft  = TRUE;	//
	if(PlayerIsRightWall() && p->Speed.x >  0) hitRight = TRUE;	//E
	if(PlayerIsUpperWall() && p->Speed.y <= 0) hitUpper = TRUE;	//

	//ǂڐG
	if(hitUpper == TRUE && hitLeft == FALSE && hitRight == FALSE)
	{
		PlayerNormalizeToUpper();
	}

	//EǂڐG
	if(hitUpper == FALSE && hitLeft == TRUE)
	{
		PlayerNormalizeToLeft();
	}

	//ǂڐG
	if(hitUpper == FALSE && hitRight == TRUE)
	{
		PlayerNormalizeToRight();
	}

	//ǂƉEǂɐڐG
	if(hitUpper == TRUE && hitRight == TRUE)
	{
		if(PlayerIsUpperWallBoth() == TRUE)
		{
			PlayerNormalizeToUpper();
		}
		else
		{
			if( PlayerToFieldOfsX() > PlayerToFieldOfsY() )
			{
				PlayerNormalizeToRight();
			}
			else
			{
				PlayerNormalizeToUpper();
			}
		}
	}

	//ǂƍǂɐڐG
	if(hitUpper == TRUE && hitLeft == TRUE)
	{
		if(PlayerIsUpperWallBoth() == TRUE)
		{
			PlayerNormalizeToUpper();
		}
		else
		{
			if( BG_SIZE - PlayerToFieldOfsX() > PlayerToFieldOfsY() )
			{
				PlayerNormalizeToLeft();
			}else{
				PlayerNormalizeToUpper();
			}
		}
	}

	//--------------------------------------------------------------------
	//
	switch( PlayerGetOnField() )
	{
	case FIELD_BG_SCROLL_L: p->Speed.x = p->Speed.x * (1.0 - PLAYER_GRD_ACCRATIO) + (p->Direction * PLAYER_SPEED - FIELD_SCROLLPANEL_SPEED) * PLAYER_GRD_ACCRATIO; break;
	case FIELD_BG_SCROLL_R: p->Speed.x = p->Speed.x * (1.0 - PLAYER_GRD_ACCRATIO) + (p->Direction * PLAYER_SPEED + FIELD_SCROLLPANEL_SPEED) * PLAYER_GRD_ACCRATIO; break;
	case FIELD_BG_NONE:     p->Speed.x = p->Speed.x * (1.0 - PLAYER_AIR_ACCRATIO) +  p->Direction * PLAYER_SPEED * PLAYER_AIR_ACCRATIO; break;
	case FIELD_BG_SLIP:     break;
	default:             p->Speed.x = p->Speed.x * (1.0 - PLAYER_GRD_ACCRATIO) +  p->Direction * PLAYER_SPEED * PLAYER_GRD_ACCRATIO; break;
	}

	//--------------------------------------------------------------------
	//r[̍XV
//	Vec2D v = gamemain.view.getPosition();
//	v.x = v.x * 0.95 + (position.x + speed.x * VIEW_DIRECTION_OFFSET) * 0.05;
//	v.y = v.y * 0.95 + position.y * 0.05;
//	gamemain.view.setPosition(v);
}
//---------------------------------------------------------------------------
void PlayerMoveItemGet()
{
	ST_PLAYER* p = &Player;

	if(p->WaitTimer < PLAYER_WAIT_TIMER_INTERVAL)
	{
		p->WaitTimer++;
		return;
	}
		
	if( PlayerIsPushEnter() == TRUE )
	{
		p->State = PLAYER_STATE_NORMAL;
	}
}
//---------------------------------------------------------------------------
void PlayerMoveByInput()
{
	ST_PLAYER* p = &Player;

	if( PlayerIsPressLeft()  == TRUE ) p->Direction = -1;
	if( PlayerIsPressRight() == TRUE ) p->Direction = 1;

	if( PlayerIsPushEnter() == TRUE  &&
	    PlayerIsPressDown() == FALSE &&
	    (p->JumpMax > p->JumpCnt || PlayerOnWall() == TRUE) )
	{
		p->Speed.y = PLAYER_JUMP;		//Wv

		if(PlayerOnWall() == FALSE)
		{
			p->JumpCnt++;
		}
	}
}
//---------------------------------------------------------------------------
//eڐG
void PlayerCheckCollision()
{
	ST_PLAYER* p = &Player;
	s16 x = PlayerToFieldX();
	s16 y = PlayerToFieldY();
	s16 xx, yy;

	for(xx=0; xx<2; xx++)
	{
		for(yy=0; yy<2; yy++)
		{
			//ACel(STATE_ITEMGET֑J)
			if( FieldIsItem(x, y) == TRUE )
			{
				//BACe͏Kv
				if(FieldIsItemGetable(x + xx, y + yy) == FALSE)
				{
					continue;
				}

				p->State   = PLAYER_STATE_ITEMGET;
				p->ItemGet = FieldGetArea(x + xx, y + yy);		//ACe
				FieldEraseArea(x + xx, y + yy);

				switch(p->ItemGet)
				{
					case FIELD_BG_ITEM_POWERUP:
						p->JumpMax++;
						break;

					default:
						p->ItemGetFlag[p->ItemGet] = TRUE;
						break;
				}
				p->WaitTimer = 0;
				return;
			}

			//gQ
			if( FieldIsSpike(x + xx, y + yy) == TRUE )
			{
				p->State      = PLAYER_STATE_MUTEKI;
				p->Life      -= PLAYER_LIFE_RATIO;
				p->Speed.y    = PLAYER_JUMP;
				p->JumpCnt    = -1;			// _[WEGLXgWv
				p->WaitTimer  =  0;

				//S
				if(p->Life < PLAYER_LIFE_RATIO)
				{
					p->State     = PLAYER_STATE_DEAD;
					p->Direction = 0;
				}
				return;
			}

		} //for xx
	} //for yy
}
//---------------------------------------------------------------------------
//Ă̂Ԃ
u8 PlayerGetOnField()
{
	if( PlayerOnWall() == FALSE )
	{
		return FIELD_BG_NONE;
	}

	if( PlayerToFieldOfsX() < BG_SIZE / 2)
	{
		if( FieldIsRidable(PlayerToFieldX(), PlayerToFieldY() + 1) == TRUE )
		{
			return FieldGetArea( PlayerToFieldX() + 0, PlayerToFieldY() + 1 );
		}
		else
		{
			return FieldGetArea( PlayerToFieldX() + 1, PlayerToFieldY() + 1 );
		}
	}
	else
	{
		if( FieldIsRidable( PlayerToFieldX() + 1, PlayerToFieldY() + 1 ) == TRUE )
		{
			return FieldGetArea( PlayerToFieldX() + 1, PlayerToFieldY() + 1 );
		}
		else
		{
			return FieldGetArea( PlayerToFieldX(), PlayerToFieldY() + 1 );
		}
	}
}
//---------------------------------------------------------------------------
//ǂɏĂ邩
bool PlayerOnWall()
{
	if( PlayerToFieldOfsY() > BG_SIZE / 4 )
	{
		return FALSE;
	}
	if( FieldIsRidable( PlayerToFieldX(), PlayerToFieldY() + 1 ) == TRUE &&
	    PlayerToFieldOfsX() < BG_SIZE * 7 / 8 )
	{
		return TRUE;
	}

	if( FieldIsRidable( PlayerToFieldX() + 1, PlayerToFieldY() + 1 ) == TRUE &&
	    PlayerToFieldOfsX() > BG_SIZE / 8)
	{
		return TRUE;
	}

	return FALSE;
}
//---------------------------------------------------------------------------
//邩
bool PlayerIsFallable()
{
	if(PlayerOnWall() == FALSE)
	{
		return FALSE;
	}

	if( FieldIsWall( PlayerToFieldX(), PlayerToFieldY() + 1 ) == TRUE && 
	    PlayerToFieldOfsX() < BG_SIZE * 7 / 8 )
	{
		return FALSE;
	}

	if( FieldIsWall( PlayerToFieldX() + 1, PlayerToFieldY() + 1 ) == TRUE &&
	    PlayerToFieldOfsX() > BG_SIZE / 8 )
	{
		return FALSE;
	}

	return TRUE;
}
//---------------------------------------------------------------------------
//mF(Q}XƂ)
bool PlayerIsUpperWallBoth()
{
	if( PlayerToFieldOfsY() < BG_SIZE / 2 )
	{
		return FALSE;
	}

	if( FieldIsWall( PlayerToFieldX() + 0, PlayerToFieldY() ) == TRUE && 
	    FieldIsWall( PlayerToFieldX() + 1, PlayerToFieldY() ) == TRUE )
	{
		return TRUE;
	}

	return FALSE;
}
//---------------------------------------------------------------------------
//mF
bool PlayerIsUpperWall()
{
	if( PlayerToFieldOfsY() < BG_SIZE / 2 )
	{
		return FALSE;
	}

	if( FieldIsWall( PlayerToFieldX(), PlayerToFieldY() == TRUE ) && 
	    PlayerToFieldOfsX() < BG_SIZE * 7 / 8 )
	{
		return TRUE;
	}

	if( FieldIsWall( PlayerToFieldX() + 1, PlayerToFieldY() == TRUE ) &&
	    PlayerToFieldOfsX() > BG_SIZE / 8 )
	{
		return TRUE;
	}

	return FALSE;
}
//---------------------------------------------------------------------------
//ǊmF
bool PlayerIsLeftWall()
{
	if( FieldIsWall( PlayerToFieldX(), PlayerToFieldY() ) == TRUE )
	{
		return FALSE;
	}

	if( FieldIsWall( PlayerToFieldX(), PlayerToFieldY() + 1 ) == TRUE &&
	    PlayerToFieldOfsY() > BG_SIZE / 8 )
	{
		return TRUE;
	}

	return FALSE;
}
//---------------------------------------------------------------------------
//EǊmF
bool PlayerIsRightWall()
{
	if( FieldIsWall( PlayerToFieldX() + 1, PlayerToFieldY() ) == TRUE )
	{
		return TRUE;
	}

	if( FieldIsWall( PlayerToFieldX() + 1, PlayerToFieldY() + 1 ) && 
	    PlayerToFieldOfsY() > BG_SIZE / 8 )
	{
		return TRUE;
	}

	return FALSE;
}
//---------------------------------------------------------------------------
//EǂɂĒ~
void PlayerNormalizeToRight()
{
	ST_PLAYER* p = &Player;

	p->Position.x = PlayerToFieldX() * BG_SIZE;
	p->Speed.x    = 0;
}
//---------------------------------------------------------------------------
//ǂɂĒ~
void PlayerNormalizeToLeft()
{
	ST_PLAYER* p = &Player;

	p->Position.x = (PlayerToFieldX() + 1) * BG_SIZE;
	p->Speed.x    = 0;
}
//---------------------------------------------------------------------------
//ǂɂĒ~
void PlayerNormalizeToUpper()
{
	ST_PLAYER* p = &Player;

	if(p->Speed.y < 0)
	{
		p->Speed.y = 0;
	}

	p->Position.y = BG_SIZE * (PlayerToFieldY() + 1);
}
//---------------------------------------------------------------------------
u16 PlayerToFieldX()
{
	return (u16)(Player.Position.x / BG_SIZE);
}
//---------------------------------------------------------------------------
u16 PlayerToFieldY()
{
	return (u16)(Player.Position.y / BG_SIZE);
}
//---------------------------------------------------------------------------
//tB[h}bvWnϊ̗]
u16 PlayerToFieldOfsX()
{
	return (u16)Player.Position.x % BG_SIZE;
}
//---------------------------------------------------------------------------
u16 PlayerToFieldOfsY()
{
	return (u16)Player.Position.y % BG_SIZE;
}
//---------------------------------------------------------------------------
void PlayerDraw()
{

}
//---------------------------------------------------------------------------
bool PlayerIsPressEnter()
{
	return Player.Key & KEY_A ? TRUE : FALSE;
}
bool PlayerIsPushEnter()
{
	return Player.Key & KEY_A ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
bool PlayerIsPressDown()
{
	return Player.Key & KEY_DOWN ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
bool PlayerIsPressLeft()
{
	return Player.Key & KEY_LEFT ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
bool PlayerIsPressRight()
{
	return Player.Key & KEY_RIGHT ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
void PlayerViewSetPosition(ST_VEC2D* p)
{
	Player.View.x = p->x;
	Player.View.y = p->y;
}
//---------------------------------------------------------------------------
void BGDraw16(u8 x, u8 y, u16 tile)
{
	BGDrawChar(BG2.Map, x*2+0, y*2+0, BG2.Col, tile+0 );
	BGDrawChar(BG2.Map, x*2+1, y*2+0, BG2.Col, tile+1 );
	BGDrawChar(BG2.Map, x*2+0, y*2+1, BG2.Col, tile+32);
	BGDrawChar(BG2.Map, x*2+1, y*2+1, BG2.Col, tile+33);
}
